/*  Light to Hight -- rising map plug-in.
 *  Copyright (C) 2001-2006 Kyoichiro Suda
 * 
 * The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Light to Height version 0.7, Mar 2007
 *
 * Changes
 *
 *  fixed coordinates at inside of selection.
 *  added preview.
 * 
 * TODO
 *
 *  shorten, mass.
 *
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#define PLUG_IN_NAME    "plug_in_light_to_height"
#define SCALE_WIDTH      0

#ifdef ENABLE_NLS
#define GETTEXT_PACKAGE "gimp20-l2height"

#include <libintl.h>
#define _(String) gettext (String)
#ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#else
#    define N_(String) (String)
#endif

/* #define bind_textdomain_codeset(Domain, Codeset) (Domain) */
#define INIT_I18N()     G_STMT_START{                           \
  bindtextdomain (GETTEXT_PACKAGE, gimp_locale_directory ());   \
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");           \
  textdomain (GETTEXT_PACKAGE);                                 \
}G_STMT_END

#else /* ! ENABLE_NLS */
#define _(String) (String)
#define N_(String) (String)
#endif


#define PREVIEW_INVALIDATE(widget,signal) g_signal_connect_swapped (widget, signal, G_CALLBACK (gimp_preview_invalidate), preview)

/*---- Types ---*/

typedef struct
{
  gint32        map_id;
  gdouble       grow;
  gdouble       shorten;
  gboolean      with_smooth;
  gboolean      with_wrap;
  gboolean      with_mass;
  gboolean      with_horizon;
  gboolean      with_shorten;
  gboolean      preview;
} L2HValues;

typedef struct
{
  gint32            drawable_id;
  GimpDrawable     *drawable;
  GimpPixelFetcher *pf;
  gint              length;
  gdouble          *data;
  gboolean          is_rgb       :1;
  gboolean          has_alpha    :1;
  gboolean          with_horizon :1;
} MapParams;

typedef struct
{
  gint           length;
  guchar        *src_buf;
  guchar        *dest_buf;
  gdouble       *map_buf; // copy of pointer from MapParams
  gdouble        grow;
  gdouble        surface;
  gint           bppa;
  gint           bpp;
  gboolean       is_rgb      :1;
  gboolean       has_alpha   :1;
  gboolean       with_smooth :1;
  gboolean       with_wrap   :1;
  gboolean       with_mass   :1;
} ImgParams;

typedef struct
{
  GimpPreview   *preview;
  guint          idle;
  guint          idle_draw;
  gint           com_line;
  gint           draw_line;
  gint           line_o;
  gint           lines;
  gint           c_line_o;
  gint           c_lines;
  gint           x, y, w, h;
  GimpPixelRgn   src_rgn;
  GimpPixelRgn   dest_rgn;
  MapParams     *map;
  ImgParams     *img;
  gdouble        shorten;
  gboolean       with_horizon :1;
  gboolean       with_shorten :1;
  gboolean       preserved    :1;
} PreviewValues;


/*---- Prottype ---*/

static void     query           (void);

static void     run             (const gchar      *name,
                                 gint              nparams,
                                 const GimpParam  *param,
                                 gint             *nreturn_vals,
                                 GimpParam       **return_vals);

static MapParams* map_params    (gint32          map_id,
                                 gint            length);
static void     map_params_free (MapParams      *map);
static void     map_get_line    (MapParams      *map,
                                 gint            line,
                                 gint            pix);
static gdouble  map_get         (MapParams      *map,
                                 gint            line,
                                 gint            pix);

static ImgParams* img_params    (GimpDrawable   *drawable,
                                 gint            length);
static void     img_params_free (ImgParams      *img);
static void     color_flush     (guchar         *dest,
                                 gdouble        *col,
                                 ImgParams      *img);
static void     l2h_compute     (ImgParams      *img);
static void     light_to_hight  (gint32          image_ID,
                                 GimpDrawable   *drawable);
static gint     l2h_dialog      (gint32          image_ID,
                                 GimpDrawable   *drawable);

static void     preview_free    (void);
static void     preview_draw    (gpointer        data);
static void     preview_progress (gpointer       data);
static void     preview_init    (GimpPreview    *preview,
                                 GimpDrawable   *drawable);
static void     preview_destroy (void);


/*---- Variables ----*/

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

static L2HValues l2h_vals =
{
  -1,           /* hight map id */
  20.0,         /* grow length */
  100.0,        /* shorten value */
  TRUE,         /* with_smooth */
  FALSE,        /* with_wrap */
  TRUE,         /* with_mass */
  FALSE,        /* with_horizon */
  FALSE,        /* use shorten */
  FALSE,        /* preview */
};

static PreviewValues l2h_preview = 
{ NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0,0,0,0,
  {}, {}, NULL, NULL, 0.0, FALSE, FALSE, FALSE };


/* Functions */

MAIN ()

static void
query (void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32,    "run_mode",    "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE,    "image",       "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable",    "Input drawable" },
    { GIMP_PDB_DRAWABLE, "map",         "Light data drawable" },
    { GIMP_PDB_FLOAT,    "grow",        "Growing max height" },
    { GIMP_PDB_INT32,    "shorten",     "Shorten map value" },
    { GIMP_PDB_INT32,    "with_smooth", "Smooth growing" },
    { GIMP_PDB_INT32,    "with_wrap",   "Wrap arownd growing" },
    { GIMP_PDB_INT32,    "with_mass",   "Grow with mass or plane" },
    { GIMP_PDB_INT32,    "with_horizon", "Turn height to horizon" },
  };

  gimp_install_procedure (PLUG_IN_NAME,
                          "Grows a image according to brightness.",
                          "This Plug-in is image mapper along height,"
                          "Image's height is measured by light.",
                          "Kyoichiro Suda <sudakyo\100fat.coara.or.jp>",
                          "Kyoichiro Suda", "2001-2005",
                          N_("_Light to Height..."),
                          "RGB*, GRAY*",
                          GIMP_PLUGIN, G_N_ELEMENTS (args), 0, args, NULL);

  gimp_plugin_menu_register (PLUG_IN_NAME, "<Image>/Filters/Map/");
}

static void
run (const gchar      *name, 
     gint              nparams, 
     const GimpParam  *param, 
     gint             *nreturn_vals,
     GimpParam       **return_vals)
{
  /* Get the runmode from the in-parameters */
  GimpRunMode run_mode = param[0].data.d_int32; 
  
  /* status variable, use it to check for errors in invocation usualy only 
     during non-interactive calling */  
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;   
  
  /* always return at least the status to the caller. */
  static GimpParam values[1];
  
  /* initialize the return of the status */     
  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
  *nreturn_vals = 1;
  *return_vals = values;

#ifdef INIT_I18N
  INIT_I18N ();
#endif

  /* get image and drawable */
  gint32 image_ID = param[1].data.d_int32;
  GimpDrawable *drawable = gimp_drawable_get (param[2].data.d_drawable);

  switch (run_mode)
    {
    case GIMP_RUN_INTERACTIVE:
      /* Possibly retrieve data */
      if (gimp_get_data_size (PLUG_IN_NAME) == sizeof l2h_vals)
        gimp_get_data (PLUG_IN_NAME, &l2h_vals);
  
      /* Get information from the dialog */
      if (! l2h_dialog (image_ID, drawable))
        return;

      break;

    case GIMP_RUN_NONINTERACTIVE:
      /* Make sure all the arguments are present */
      if (nparams != 10)
        {
          status = GIMP_PDB_CALLING_ERROR;
        }
      else
        {
          l2h_vals.map_id       = (gint32) param[3].data.d_drawable;
          l2h_vals.grow         = MAX (param[4].data.d_float, 0.0);
          l2h_vals.shorten      = CLAMP (param[5].data.d_float, 0.0, 100.0);
          l2h_vals.with_smooth  = param[6].data.d_int32 ? TRUE : FALSE;
          l2h_vals.with_wrap    = param[7].data.d_int32 ? TRUE : FALSE;
          l2h_vals.with_mass    = param[8].data.d_int32 ? TRUE : FALSE;
          l2h_vals.with_horizon = param[9].data.d_int32 ? TRUE : FALSE;
          l2h_vals.with_shorten = param[5].data.d_float >= 0 ? TRUE : FALSE;
        }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /* Possibly retrieve data */
      if (gimp_get_data_size (PLUG_IN_NAME) != sizeof l2h_vals ||
          ! gimp_get_data (PLUG_IN_NAME, &l2h_vals))
        status = GIMP_PDB_CALLING_ERROR;
      break;

    default:
      break;
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      if ((gimp_drawable_is_rgb(drawable->drawable_id) ||
           gimp_drawable_is_gray(drawable->drawable_id)))
        {
          /* Run! */
          light_to_hight (image_ID, drawable);

          /* If run mode is interactive, flush displays */
          if (run_mode != GIMP_RUN_NONINTERACTIVE)
            gimp_displays_flush ();

          /* Store data */
          if (run_mode == GIMP_RUN_INTERACTIVE)
            gimp_set_data (name, &l2h_vals, sizeof l2h_vals);
        }
    }
  else
    status = GIMP_PDB_EXECUTION_ERROR;

  values[0].data.d_status = status;
  gimp_drawable_detach (drawable);
}


static MapParams*
map_params (gint32   map_id,
            gint     length)
{
  MapParams *map;
  gsize      buf_len, s_len;

  buf_len = sizeof (gdouble) * length;
  s_len   = sizeof (MapParams);
  map     = g_malloc (s_len + 31 + buf_len);
  map->data   = (gdouble *) (((gsize) map + s_len + 31) & ~31);
  map->length = length;

  map->drawable_id = map_id;
  map->drawable    = gimp_drawable_get (map_id);
  map->pf          = gimp_pixel_fetcher_new (map->drawable, FALSE);
  gimp_pixel_fetcher_set_edge_mode (map->pf, GIMP_PIXEL_FETCHER_EDGE_WRAP);

  map->is_rgb       = gimp_drawable_is_rgb (map_id);
  map->has_alpha    = gimp_drawable_has_alpha (map_id);
  map->with_horizon = l2h_vals.with_horizon ? TRUE : FALSE;

  return map;
}


static void
map_params_free (MapParams* map)
{
  if (! map)
    return;

  gimp_pixel_fetcher_destroy (map->pf);
  gimp_drawable_detach (map->drawable);

  g_free (map);
}


static void
map_get_line (MapParams *map,
              gint       line,
              gint       pix)
{
  gdouble *bp, *ep;

  bp = map->data;
  ep = bp + map->length;

  while (bp < ep)
    *(bp ++) = map_get (map, line, pix ++);
}
         

static gdouble
map_get (MapParams *map,
         gint       line,
         gint       pix)
{
  guchar  buf[8];
  gdouble val;

  if (map->with_horizon)
    gimp_pixel_fetcher_get_pixel (map->pf, pix, line, buf);
  else
    gimp_pixel_fetcher_get_pixel (map->pf, line, pix, buf);

  if (map->is_rgb)
    {
      val = GIMP_RGB_INTENSITY (buf[0], buf[1], buf[2]);
      if (map->has_alpha)
        val = val * (gdouble) buf[3] / 255;
    }
  else
    {
      val = buf[0];
      if (map->has_alpha)
        val = val * (gdouble) buf[1] / 255;
    }

  val /= 255;

  return val;
}
         

static ImgParams*
img_params (GimpDrawable *drawable,
            gint          length)
{
  ImgParams *img;
  gsize      s_len, b_len;

  s_len = sizeof (ImgParams);
  b_len = (sizeof (guchar) * drawable->bpp * length + 31) & ~31;
  img = g_malloc (s_len + 31 + b_len + b_len);
  img->src_buf  = (guchar *) (((gsize) img + s_len + 31) & ~31);
  img->dest_buf = (guchar *)  ((gsize) img->src_buf + b_len);
  img->map_buf  = NULL;
  img->length   = length;
  img->grow     = l2h_vals.grow;

  if (l2h_vals.with_shorten && ! l2h_vals.with_wrap)
    {
      gdouble    surface;
      surface = l2h_vals.shorten / 100;
      img->surface = CLAMP (surface, 0.0, 1.0);
    }
  else
    img->surface = 1.0;

  img->bppa        = drawable->bpp;
  img->bpp         = drawable->bpp;
  img->is_rgb      = gimp_drawable_is_rgb (drawable->drawable_id);
  img->has_alpha   = gimp_drawable_has_alpha (drawable->drawable_id);
  if (img->has_alpha)
    img->bpp --;

  img->with_smooth = l2h_vals.with_smooth ? TRUE : FALSE;
  img->with_wrap   = l2h_vals.with_wrap ? TRUE : FALSE;
  img->with_mass   = l2h_vals.with_mass ? TRUE : FALSE;

  return img;
}


static void
img_params_free (ImgParams *img)
{
  if (! img)
    return;

  g_free (img);
}


static void
color_flush (guchar     *dest,
             gdouble    *col,
             ImgParams  *img)
{
  gdouble a, o;
  gint    c, b;

  a = col[img->bpp];
  a = CLAMP (a, 0.0, 1.0);

  if (img->has_alpha)
    {
      if (a > 0)
        {
          for (b = 0; b < img->bpp; b ++)
            {
              c = RINT (col[b] / a);
              dest[b] = CLAMP0255 (c);
            }
          dest[img->bpp] = RINT (255 * a);
        }
      else
        {
          for (b = 0; b < img->bpp; b ++)
            dest[b] = 0;
        }
    }
  else
    {
      if (a < 1)
        {
          static gdouble col[4], gray;
          static gboolean got_bg = FALSE;
          if (! got_bg)
            {
              GimpRGB bg;
              gimp_context_get_background (&bg);
              col[0] = 255 * bg.r;
              col[1] = 255 * bg.g;
              col[2] = 255 * bg.b;
              gray  = GIMP_RGB_INTENSITY (bg.r, bg.g, bg.b);
              got_bg = TRUE;
            }
          o = 1.0 - a;
          if (img->is_rgb)
            {
              c = RINT (col[0] + col[0] * o);
              dest[0] = CLAMP0255 (c);
              c = RINT (col[1] + col[1] * o);
              dest[1] = CLAMP0255 (c);
              c = RINT (col[2] + col[2] * o);
              dest[2] = CLAMP0255 (c);
            }
          else
            {
              c = RINT (col[0] + gray * o);
              dest[0] = CLAMP0255 (c);
            }
        }
      else
        {
          for (b = 0; b < img->bpp; b ++)
            {
              c = RINT (col[b]);
              dest[b] = CLAMP0255 (c);
            }
        }
    }
}


static void
l2h_compute (ImgParams  *img)
{
  guchar       *lp, *bp, *np;   
  gint          i, j, k, n, b;
  gdouble       a, capa, fa;
  gdouble       fb, sb, xb;
  gdouble       dist, vol, high, reach;
  gdouble       nvol, nhigh, wide;
  gdouble       sum[4];

  g_assert (img->map_buf);

  a = 1.0; /* ! has_alpha */
  lp = img->dest_buf;
  for (i = 0; i < img->length; i ++, lp += img->bppa)
    {
      /* initialize */
      capa = 1;
      for (b = 0; b < img->bppa; b ++)
        {
          sum[b] = 0;
          lp[b] = 0;
        }

      /* grow up */
      j = ceil (i + img->grow);
      if (! img->with_wrap)
        j = MIN (j, img->length - 1);
      for (; j >= i; j --)
        {
          /* wrap around */
          k = j;
          while (k >= img->length) /* % is very slow with my processor */
            k -= img->length;

          /* reached length */
          dist = j - i;
          vol  = img->map_buf[k];
          high = vol * img->grow;
          reach = capa + high - dist;
          if (reach <= 0)
            continue;

          bp = &(img->src_buf[k * img->bppa]);
          /* alpha */
          if (img->has_alpha)
            a = (gdouble) bp[img->bpp] / 255;

          /* check to interpolate */
          if (img->with_smooth && reach > 1 &&
              (img->with_wrap || k < img->length - 2))
            {
              n = k + 1;

              while (n >= img->length)
                n -= img->length;
              np = &(img->src_buf[n * img->bppa]);

              /* between next */
              nvol  = img->map_buf[n];
              nhigh = nvol * img->grow;
              wide  = high - nhigh;

              /*
               *  "with_smooth" interpolates between two heights.
               *  square of pixel will be kept. if next pixel grows
               *  more than current pixel, next pixel hides width.
               */
              //g_assert (wide > 0);
            }
          else
            wide = 0;
          if (wide > 0)
            {
              fb = reach - 1 / capa;
              fb = CLAMP (fb, 0, 1);

              sb = (reach - 1 - capa / 2) / wide;
              sb = CLAMP (sb, 0, 1);

              xb = fb * sb;

              if (img->has_alpha)
                {
                  gdouble na, b0, b1;

                  na = (gdouble) np[img->bpp] / 255;

                  b0 = (1 - xb) * a;
                  b1 = xb * na;

                  for (b = 0; b < img->bpp; b ++)
                    sum[b] += (bp[b] * b0 + np[b] * b1) * capa;
                  fa = (b0 + b1) * capa;
                  sum[img->bpp] += fa;
                  if (img->with_mass)
                    capa -= fa;
                  else
                    capa = 0;
                }
              else /* ! has_alpha */
                {
                  for (b = 0; b < img->bpp; b ++)
                    sum[b] += ((1 - sb) * bp[b] + sb * np[b]) * capa;
                  sum[img->bpp] = 1;
                  capa = 0;
                  break;
                }
            }

          else if (a == 0 && img->with_mass)
            continue;

          else if (! img->with_smooth && reach > 0)
            {
              if (img->with_mass)
                {
                  fa = a * capa;
                  for (b = 0; b < img->bpp; b ++)
                    sum[b] += bp[b] * fa;
                  sum[img->bpp] += fa;
                  capa -= fa;
                }
              else
                {
                  for (b = 0; b < img->bpp; b ++)
                    sum[b] = bp[b] * a;
                  sum[img->bpp] = a;
                  capa = 0;
                }
            }

          else if (reach >= capa)
            {
              fa = a * capa;
              /* full covered */
              for (b = 0; b < img->bpp; b ++)
                sum[b] += (gdouble) bp[b] * fa;
              sum[img->bpp] += fa;
              if (img->with_mass)
                capa -= fa;
              else
                capa = 0;
            }

          else /* if (reach > 0 && reach < capa) */
            {
              fa = a * reach;
              /* reached */
              for (b = 0; b < img->bpp; b ++)
                sum[b] += (gdouble) bp[b] * fa;
              sum[img->bpp] += fa;
              if (img->with_mass)
                capa -= fa;
              else
                capa -= reach;
            }

          if (capa <= 0)
            break;
        }

      if (capa > 0)
        {
          /* needs shorten but stil not be writen */
          bp = &(img->src_buf[i * img->bppa]);

          if (img->has_alpha)
            {
              a = (gdouble) bp[img->bpp] / 255;
              fa = a * capa;
              for (b = 0; b < img->bpp; b ++)
                sum[b] += bp[b] * fa;
              sum[img->bpp] += fa;
              if (img->with_mass)
                capa -= fa;
              else
                capa = 0;
            }
          else
            {
              for (b = 0; b < img->bpp; b ++)
                sum[b] += bp[b] * capa;
              sum[img->bpp] += capa;
              capa = 0;
            }
        }

      color_flush (lp, sum, img);
    }
}


static void
light_to_hight (gint32        image_ID,
                GimpDrawable *drawable)
{
  ImgParams     *img;
  MapParams     *map;
  GimpPixelRgn   src_rgn, dest_rgn;
  gint           sel_x, sel_y, sel_width, sel_height;
  gint           ta, tb, tc;
  gint           len, nlines, pix_o;
  gint           i,  x, y, w, h;
  gint          *linep;

  if (l2h_vals.grow <= 0.0)
    return;

  /*  Get selection area  */
  gimp_drawable_mask_intersect (drawable->drawable_id,
                                &sel_x, &sel_y, &sel_width, &sel_height);
  if (l2h_vals.with_horizon)
    {
      ta = sel_width;
      tb = gimp_tile_width ();
    }
  else
    {
      ta = sel_height;
      tb = gimp_tile_height ();
    }
  tc = 2 * ((ta + tb - 1) / tb);
  gimp_tile_cache_ntiles (tc);

  gimp_pixel_rgn_init (&src_rgn, drawable,
                       sel_x, sel_y, sel_width, sel_height, FALSE, FALSE);
  gimp_pixel_rgn_init (&dest_rgn, drawable,
                       sel_x, sel_y, sel_width, sel_height, TRUE, TRUE);

  x = sel_x;
  y = sel_y;
  w = sel_width;
  h = sel_height;

  if (l2h_vals.with_horizon)
    {
      nlines = sel_height;
      len    = sel_width;

      pix_o  = sel_x;
      h      = 1;
      linep  = &y;
    }
  else
    {
      nlines = sel_width;
      len    = sel_height;

      pix_o  = sel_y;
      w      = 1;
      linep  = &x;
    }

  map = map_params (l2h_vals.map_id, len);
  img = img_params (drawable, len);
  img->map_buf = map->data;

  gimp_progress_init ("Light to Height..");
  for (i = 0; i < nlines; i ++, (*linep) ++)
    {
      map_get_line (map, *linep, pix_o);

      gimp_pixel_rgn_get_rect (&src_rgn,  img->src_buf,  x, y, w, h);
      l2h_compute (img);
      gimp_pixel_rgn_set_rect (&dest_rgn, img->dest_buf, x, y, w, h);

      if ((i & 63) == 0)
        gimp_progress_update ((double) (i + 1) / (double) nlines);
    }

  img_params_free (img);
  map_params_free (map);

  /* update */
  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
  gimp_drawable_update (drawable->drawable_id,
                        sel_x, sel_y, sel_width, sel_height);

  return;
}


static gint
mapmenu_constraint (gint32   image_id,
                    gint32   drawable_id,
                    gpointer data)
{
  if (drawable_id == -1)
    return TRUE;

  return (gimp_drawable_is_rgb (drawable_id) ||
          gimp_drawable_is_gray (drawable_id));
}


static gint
l2h_dialog (gint32        image_ID,
            GimpDrawable *drawable)
{
  GtkWidget *dialog;
  GtkWidget *main_vbox;
  GtkWidget *preview;

  GtkWidget *table;
  GtkWidget *combo;
  GtkWidget *toggle;
  GtkObject *adj;
  gboolean   run;

  gimp_ui_init ("l2height", TRUE);

  dialog = gimp_dialog_new (_("Light to Height"), "light_to_height",
                            NULL, 0,
                            gimp_standard_help_func, NULL,
                            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                            GTK_STOCK_OK,     GTK_RESPONSE_OK,
                            NULL);
  gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

  main_vbox = gtk_vbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (main_vbox), 12);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), main_vbox);
  gtk_widget_show (main_vbox);

  /* preview */
  preview = gimp_drawable_preview_new (drawable, &l2h_vals.preview);
  gtk_box_pack_start (GTK_BOX (main_vbox), preview, TRUE, TRUE, 0);
  gtk_widget_show (preview);
  g_signal_connect (preview, "invalidated",
                    G_CALLBACK (preview_init), drawable);

  table = gtk_table_new (4, 4, FALSE);
  gtk_table_set_col_spacings (GTK_TABLE (table), 6);
  gtk_table_set_row_spacings (GTK_TABLE (table), 6);
  gtk_box_pack_start (GTK_BOX (main_vbox), table, FALSE, FALSE, 0);
  gtk_widget_show (table);

  /* smooth */
  toggle = gtk_check_button_new_with_mnemonic (_("S_mooth"));
  gimp_help_set_help_data (toggle, _("Interpolates between height points."), NULL);
  gtk_table_attach (GTK_TABLE (table), toggle, 3, 4, 0, 1,
                    GTK_SHRINK | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
                                l2h_vals.with_smooth);
  g_signal_connect (toggle, "toggled",
                    G_CALLBACK (gimp_toggle_button_update),
                    &l2h_vals.with_smooth);
  PREVIEW_INVALIDATE (toggle, "toggled");
  gtk_widget_show (toggle);

  /* wrap around */
  toggle = gtk_check_button_new_with_mnemonic (_("_Wrap"));
  gimp_help_set_help_data (toggle, _("Wraps around growing."), NULL);
  gtk_table_attach (GTK_TABLE (table), toggle, 3, 4, 1, 2,
                    GTK_SHRINK | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
                                l2h_vals.with_wrap);
  g_signal_connect (toggle, "toggled",
                    G_CALLBACK (gimp_toggle_button_update),
                    &l2h_vals.with_wrap);
  PREVIEW_INVALIDATE (toggle, "toggled");
  gtk_widget_show (toggle);

  /* turn to horizon */
  toggle = gtk_check_button_new_with_mnemonic (_("Ho_rizon"));
  gimp_help_set_help_data (toggle, _("Grows to horizon."), NULL);
  gtk_table_attach (GTK_TABLE (table), toggle, 3, 4, 2, 3,
                    GTK_SHRINK | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
                                l2h_vals.with_horizon);
  g_signal_connect (toggle, "toggled",
                    G_CALLBACK (gimp_toggle_button_update),
                    &l2h_vals.with_horizon);
  PREVIEW_INVALIDATE (toggle, "toggled");
  gtk_widget_show (toggle);

  /* shorten map flag */
  /*
  toggle = gtk_check_button_new_with_mnemonic (_("Shor_ten"));
  gimp_help_set_help_data (toggle, _("Shortens rendered image."), NULL);
  gtk_table_attach (GTK_TABLE (table), toggle, 3, 4, 3, 4,
                    GTK_SHRINK | GTK_FILL, GTK_FILL, 1, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle),
                                l2h_vals.with_shorten);
  g_signal_connect (toggle, "toggled",
                    G_CALLBACK (gimp_toggle_button_update),
                    &l2h_vals.with_shorten);
  PREVIEW_INVALIDATE (toggle, "toggled");
  gtk_widget_show (toggle);
  */

  /* map */
  combo = gimp_drawable_combo_box_new (mapmenu_constraint, NULL);
  gimp_int_combo_box_connect (GIMP_INT_COMBO_BOX (combo),
                              l2h_vals.map_id,
                              G_CALLBACK (gimp_int_combo_box_get_active),
                              &l2h_vals.map_id);
  gimp_table_attach_aligned (GTK_TABLE (table), 0, 1,
                             _("_Light:"), 1.0, 0.5,
                             combo, 2, TRUE);
  gimp_help_set_help_data (combo, _("Light data."), NULL);
  PREVIEW_INVALIDATE (combo, "changed");
  gtk_widget_show (combo);

  /* height */
  adj = gimp_scale_entry_new (GTK_TABLE (table), 0, 2,
                              _("_Height:"), SCALE_WIDTH, 0,
                              l2h_vals.grow, 0.0, 500.0, 1.0, 10.0, 2,
                              TRUE, 0, 0,
                              _("Height for growing."), NULL);
  g_signal_connect (adj, "value_changed",
                    G_CALLBACK (gimp_double_adjustment_update),
                    &l2h_vals.grow);
  PREVIEW_INVALIDATE (adj, "value_changed");

  /* rate to shorten map */
  /*
  adj = gimp_scale_entry_new (GTK_TABLE (table), 0, 3,
                              _("_Shorten:"), SCALE_WIDTH, 0,
                              l2h_vals.shorten, 0.0, 100.0, 1, 10, 1,
                              TRUE, 0, 0,
                              _("Rate for shotening."), NULL);
  g_signal_connect (adj, "value_changed",
                    G_CALLBACK (gimp_double_adjustment_update),
                    &l2h_vals.shorten);
  PREVIEW_INVALIDATE (adj, "value_changed");
  */

  /* finish */
  gtk_widget_show_all (dialog);

  run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);
  preview_destroy ();
  gtk_widget_destroy (dialog);

  return run;
}


/*---- Preview ----*/

static void
preview_free (void)
{
  if (l2h_preview.idle)
    {
      g_source_remove (l2h_preview.idle);
      l2h_preview.idle = 0;
    }

  if (l2h_preview.idle_draw)
    {
      g_source_remove (l2h_preview.idle_draw);
      l2h_preview.idle = 0;
    }

  /* preserve passed lines */
  if (l2h_preview.com_line > 0)
    {
      if (! l2h_preview.preserved)
        {
          l2h_preview.c_line_o  = l2h_preview.line_o;
          l2h_preview.c_lines   = l2h_preview.com_line;
          l2h_preview.preserved = TRUE;
        }

      else if (l2h_preview.line_o < l2h_preview.c_line_o)
        {
          if (l2h_preview.com_line == l2h_preview.lines)
            {
              l2h_preview.c_line_o  = l2h_preview.line_o;
              l2h_preview.c_lines  += l2h_preview.com_line;
              l2h_preview.preserved = TRUE;
            }
          else if (l2h_preview.com_line == l2h_preview.lines ||
                   l2h_preview.com_line > (l2h_preview.c_lines >> 1))
            {
              l2h_preview.c_line_o  = l2h_preview.line_o;
              l2h_preview.c_lines   = l2h_preview.com_line;
              l2h_preview.preserved = TRUE;
            }
        }
      else
        {
          if (l2h_preview.line_o ==
              l2h_preview.c_line_o + l2h_preview.c_lines)
            {
              l2h_preview.c_lines  += l2h_preview.com_line;
              l2h_preview.preserved = TRUE;
            }
          else if (l2h_preview.com_line == l2h_preview.lines ||
                   l2h_preview.com_line > (l2h_preview.c_lines >> 1))
            {
              l2h_preview.c_line_o  = l2h_preview.line_o;
              l2h_preview.c_lines   = l2h_preview.com_line;
              l2h_preview.preserved = TRUE;
            }
        }
    }

  l2h_preview.com_line = 0;
  l2h_preview.draw_line = 0;
  l2h_preview.lines = 0;
}


static void
preview_draw (gpointer data)
{
  g_assert (l2h_preview.lines > 0);

  if (l2h_preview.com_line == l2h_preview.draw_line)
    return;

  gimp_drawable_preview_draw_region
    (GIMP_DRAWABLE_PREVIEW (l2h_preview.preview), &l2h_preview.dest_rgn);

  l2h_preview.draw_line = l2h_preview.com_line;
}


static void
preview_progress (gpointer data)
{
  gint line, pix_o;
  gint x, y, w, h;

  g_assert (l2h_preview.lines > 0);
  g_assert (l2h_preview.com_line < l2h_preview.lines);

  line = l2h_preview.line_o + l2h_preview.com_line;
  if (l2h_preview.with_horizon)
    {
      x = l2h_preview.x;
      y = line;
      w = l2h_preview.w;
      h = 1;
      pix_o = x;
    }
  else
    {
      x = line;
      y = l2h_preview.y;
      w = 1;
      h = l2h_preview.h;
      pix_o = y;
    }

  map_get_line (l2h_preview.map, line, pix_o);
  gimp_pixel_rgn_get_rect (&l2h_preview.src_rgn,
                           l2h_preview.img->src_buf,  x, y, w, h);
  l2h_compute (l2h_preview.img);
  gimp_pixel_rgn_set_rect (&l2h_preview.dest_rgn,
                           l2h_preview.img->dest_buf, x, y, w, h);

  l2h_preview.com_line ++;

  /* complete */
  if (l2h_preview.com_line == l2h_preview.lines)
    {
      gimp_drawable_preview_draw_region
        (GIMP_DRAWABLE_PREVIEW (l2h_preview.preview), &l2h_preview.dest_rgn);
      preview_free ();
    }
}


static void
preview_init (GimpPreview  *preview,
              GimpDrawable *drawable)
{
  gint px, py, pw, ph;
  gint lines, line_o, length;

  preview_free ();

  if (l2h_vals.grow <= 0.0)
    return;

  gimp_preview_get_position (preview, &px, &py);
  gimp_preview_get_size (preview, &pw, &ph);


  /* first time */
  if (! l2h_preview.preview)
    {
      gint tw, th, tx, ty;
      gimp_drawable_mask_intersect (drawable->drawable_id,
                                    &l2h_preview.x, &l2h_preview.y,
                                    &l2h_preview.w, &l2h_preview.h);
      tw = gimp_tile_width ();
      th = gimp_tile_height ();
      tx = (l2h_preview.w + tw - 1) / tw;
      ty = (l2h_preview.h + th - 1) / th;
      gimp_tile_cache_ntiles (MAX (tx, ty) * 2);

      gimp_pixel_rgn_init (&l2h_preview.src_rgn, drawable,
                           l2h_preview.x, l2h_preview.y,
                           l2h_preview.w, l2h_preview.h, FALSE, FALSE);
      gimp_pixel_rgn_init (&l2h_preview.dest_rgn, drawable,
                           l2h_preview.x, l2h_preview.y,
                           l2h_preview.w, l2h_preview.h, FALSE, TRUE);

      length = MAX (l2h_preview.w, l2h_preview.h);
      l2h_preview.map = map_params (l2h_vals.map_id, length);
      l2h_preview.img = img_params (drawable, length);
      l2h_preview.img->map_buf = l2h_preview.map->data;

      if (l2h_vals.with_horizon)
        {
          l2h_preview.map->with_horizon = TRUE;
          l2h_preview.map->length = l2h_preview.w;
          l2h_preview.img->length = l2h_preview.w;
        }
      else
        {
          l2h_preview.map->with_horizon = FALSE;
          l2h_preview.map->length = l2h_preview.h;
          l2h_preview.img->length = l2h_preview.h;
        }

      l2h_preview.shorten = l2h_vals.shorten;
      l2h_preview.with_horizon = l2h_vals.with_horizon ? TRUE : FALSE;
      l2h_preview.with_shorten = l2h_vals.with_shorten ? TRUE : FALSE;
      l2h_preview.preview = preview;
      l2h_preview.preserved = FALSE;
    }

  else
    {
      /* parameter has been changed */
      if (l2h_vals.grow         != l2h_preview.img->grow        ||
          l2h_vals.shorten      != l2h_preview.shorten          ||
          (l2h_vals.with_smooth ? TRUE : FALSE) !=
          (l2h_preview.img->with_smooth ? TRUE : FALSE)         ||
          (l2h_vals.with_wrap ? TRUE : FALSE)   !=
          (l2h_preview.img->with_wrap ? TRUE : FALSE)           ||
          (l2h_vals.with_mass ? TRUE : FALSE)   !=
          (l2h_preview.img->with_mass ? TRUE : FALSE)           ||
          (l2h_vals.with_horizon ? TRUE : FALSE)        !=
          (l2h_preview.with_horizon ? TRUE : FALSE)             ||
          (l2h_vals.with_shorten ? TRUE : FALSE)        !=
          (l2h_preview.with_shorten ? TRUE : FALSE))
        {
          l2h_preview.img->grow         = l2h_vals.grow;
          l2h_preview.shorten           = l2h_vals.shorten;
          l2h_preview.img->with_smooth  = l2h_vals.with_smooth ? TRUE : FALSE;
          l2h_preview.img->with_wrap    = l2h_vals.with_wrap ? TRUE : FALSE;
          l2h_preview.img->with_mass    = l2h_vals.with_mass ? TRUE : FALSE;
          l2h_preview.with_horizon      = l2h_vals.with_horizon ? TRUE : FALSE;
          l2h_preview.with_shorten      = l2h_vals.with_shorten ? TRUE : FALSE;
          if (l2h_vals.with_horizon)
            {
              l2h_preview.map->with_horizon = TRUE;
              l2h_preview.map->length = l2h_preview.w;
              l2h_preview.img->length = l2h_preview.w;
            }
          else
            {
              l2h_preview.map->with_horizon = FALSE;
              l2h_preview.map->length = l2h_preview.h;
              l2h_preview.img->length = l2h_preview.h;
            }
          if (l2h_vals.with_shorten && ! l2h_vals.with_wrap)
            {
              gdouble    surface;
              surface = l2h_vals.shorten / 100;
              l2h_preview.img->surface = CLAMP (surface, 0.0, 1.0);
            }
          else
            l2h_preview.img->surface = 1.0;
          l2h_preview.preserved = FALSE;
        }

      /* another map was selected */
      if (l2h_vals.map_id != l2h_preview.map->drawable_id)
        {
          map_params_free (l2h_preview.map);
          length = MAX (l2h_preview.w, l2h_preview.h);
          l2h_preview.map = map_params (l2h_vals.map_id, length);
          l2h_preview.img->map_buf = l2h_preview.map->data;
          if (l2h_vals.with_horizon)
            {
              l2h_preview.map->with_horizon = TRUE;
              l2h_preview.map->length = l2h_preview.w;
            }
          else
            {
              l2h_preview.map->with_horizon = FALSE;
              l2h_preview.map->length = l2h_preview.h;
            }
          l2h_preview.preserved = FALSE;
        }
    }

  if (l2h_preview.with_horizon)
    {
      lines  = ph;
      length = l2h_preview.w;
      line_o = py;
    }
  else
    {
      lines  = pw;
      length = l2h_preview.h;
      line_o = px;
    }

  if (! l2h_preview.preserved)
    goto preview_entry;


  /* preserve passed lines */
  if (line_o >= l2h_preview.c_line_o &&
      line_o + lines <= l2h_preview.c_line_o + l2h_preview.c_lines)
    {
      gimp_drawable_preview_draw_region
        (GIMP_DRAWABLE_PREVIEW (l2h_preview.preview), &l2h_preview.dest_rgn);
      return;
    }

  if (l2h_preview.c_line_o > line_o &&
      l2h_preview.c_line_o < line_o + lines)
    {
      lines = l2h_preview.c_line_o - line_o;
    }
  else if (line_o > l2h_preview.c_line_o &&
           line_o < l2h_preview.c_line_o + l2h_preview.c_lines)
    {
      lines += line_o;
      line_o = l2h_preview.c_line_o + l2h_preview.c_lines;
      lines -= line_o;
    }

 preview_entry:

  g_assert (line_o >= 0);
  g_assert (lines > 0);

  /* draw previous image */
  {
    guchar *buf;
    gint x, y, w, h;

    buf = g_malloc (drawable->bpp * lines * length);
    if (l2h_preview.with_horizon)
      {
        x = px;
        y = line_o;
        w = pw;
        h = lines;
      }
    else
      {
        x = line_o;
        y = py;
        w = lines;
        h = ph;
      }

    gimp_pixel_rgn_get_rect (&l2h_preview.src_rgn,  buf, x, y, w, h);
    gimp_pixel_rgn_set_rect (&l2h_preview.dest_rgn, buf, x, y, w, h);

    g_free (buf);
  }

  l2h_preview.line_o    = line_o;
  l2h_preview.lines     = lines;

  l2h_preview.com_line  = 0;
  l2h_preview.draw_line = 0;

  l2h_preview.idle = g_idle_add_full
    (G_PRIORITY_LOW, (GSourceFunc) preview_progress, NULL, NULL);

  l2h_preview.idle_draw = g_timeout_add_full
    (G_PRIORITY_LOW, 250, (GSourceFunc) preview_draw, NULL, NULL);
}


static void
preview_destroy (void)
{
  preview_free ();

  if (l2h_preview.img)
    {
      img_params_free (l2h_preview.img);
      l2h_preview.img = NULL;
    }

  if (l2h_preview.map)
    {
      map_params_free (l2h_preview.map);
      l2h_preview.map = NULL;
    }

  memset (&l2h_preview, 0, sizeof l2h_preview);
}
